/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northestern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: XIAO Tong (xiaotong@mail.neu.edu.cn) 2018-07-10
 */

#include <stdio.h>
#include "XNet.h"
#include "../tensor/XUtility.h"
#include "../tensor/function/FHeader.h"
#include "../tensor/core/CHeader.h"
#include "../tensor/test/Test.h"
#include "../sample/fnnlm/FNNLM.h"
#include "../sample/transformer/Transformer.h"

//#define CRTDBG_MAP_ALLOC
//#include <stdlib.h>
//#include <crtdbg.h>

void BackwardTest();
void TransposeTest();
void SumDimTest();
void SplitBackwardTest();
void MemTest();
void xcTest();
void ConvertDataTypeTest();
void ConvertDataTypeBackwardTest();
void SumFP16Test();
void GatherFP16Test();
void HardTanHFP16Test();
void ReduceMaxFP16Test();
void ReduceSumFP16Test();
void LogSoftmaxFP16Test();
void ClipFP16Test();
void ScaleAndShiftFP16Test();
void InitTensorFP16Test();

void MultiplyDimTime();
void TimeTestGemm();
void TimeTest();
void TimeInt8AndFloat32();
void TestCPUhalf();


using namespace nts;
using namespace fnnlm;
using namespace transformer;

int main(int argc, const char ** argv )
{
    //xcTest();
    //return 0;
    //MemTest();
    //return 0;
    //SplitBackwardTest();
    //return 0;
    //_CrtSetBreakAlloc(896);
    //BackwardTest();
    //return 0;
	//Test();
	//return 0;

    //ConvertDataTypeTest();
    //return 0;  
    //ConvertDataTypeBackwardTest();
    //return 0;
    //SumFP16Test();
    //return 0;
    //GatherFP16Test();
    //return 0;
    //HardTanHFP16Test();
    //return 0;
    //ReduceMaxFP16Test();
    //return 0;
    //ReduceSumFP16Test();
    //return 0;
    //LogSoftmaxFP16Test();
    //return 0;
    //ClipFP16Test();
    //return 0;
    //ScaleAndShiftFP16Test();
    //return 0;
    //InitTensorFP16Test();
    //return 0;

	if (argc > 1 && !strcmp(argv[1], "-test"))
		Test();
    else if(argc > 1 && !strcmp(argv[1], "-fnnlm"))
        FNNLMMain(argc - 1, argv + 1);
    else if(argc > 1 && !strcmp(argv[1], "-t2t"))
        TransformerMain(argc - 1, argv + 1);
    else{
        fprintf(stderr, "Thanks for using NiuTrans.Network! This is a library for building\n");
        fprintf(stderr, "neural networks in an easy way. \n\n");
        fprintf(stderr, "Run this program with \"-test\" for unit test!\n");
        fprintf(stderr, "Or run this program with \"-fnnlm\" for sample FNNLM!\n");
    }

    //_CrtDumpMemoryLeaks();
    
    return 0;
}

void TestCPUhalf() {
	int memSize = 1024;
	int devId = 0;
	int dim1 = 1024;
	int dim2 = 32;
	XMem * mem;
	mem = new XMem(devId, FREE_ON_THE_FLY, (MTYPE)MILLION * 256, 1024, MILLION * 128);
	mem->SetDesiredSize(devId, 0, (MTYPE)memSize * MILLION);

	XTensor a;
	XTensor b;
	XTensor c;

	//XMem *mem = new XMem(devId, FREE_ON_THE_FLY, 128 * MILLION, 1024, 128 * MILLION);
	//mem->SetDesiredSize(0,0,memSize*MILLION);

	InitTensor2D(&a, dim1, dim1, X_FLOAT, devId);
	InitTensor2D(&b, dim2, dim2, X_FLOAT, devId);
	InitTensor2D(&c, dim1, dim1, X_FLOAT, devId);

}

void TimeInt8AndFloat32() {
	XMem * mem;
	int memSize = 1024;
	int devId = 2;
	int dim = 512;
	mem = new XMem(devId, FREE_ON_THE_FLY, (MTYPE)MILLION * 256, 1024, MILLION * 128);
	mem->SetDesiredSize(devId, 0, (MTYPE)memSize * MILLION);

	XTensor a;
	XTensor b;
	XTensor c;
	InitTensor2D(&a, dim, dim, X_FLOAT, devId, mem);
	InitTensor2D(&b, dim, dim, X_FLOAT, devId, mem);
	InitTensor2D(&c, dim, dim, X_FLOAT, devId, mem);
	a.SetDataRand(-1.0F, 1.0F);
	b.SetDataRand(-1.0F, 1.0F);

	XTensor inta;
	XTensor intb;
	XTensor intc;
	InitTensor2D(&inta, dim, dim, X_INT, devId, mem);
	InitTensor2D(&intb, dim, dim, X_INT, devId, mem);
	InitTensor2D(&intc, dim, dim, X_FLOAT, devId, mem);

	XTensor tmp;
	InitTensor2D(&tmp, dim, dim, X_FLOAT, devId, mem);
	tmp.SetDataRand(-100000.0F, 100000.0F);
	inta = ConvertDataType(tmp, X_INT8);
	intb = ConvertDataType(tmp, X_INT8);

	int repeat = 10000;
	printf("test on matrixmul\n");
	double start_matrixmul32 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		_MatrixMul(&a, X_NOTRANS, &b, X_NOTRANS, &c);
	}
	double elapsed_matrixmul32 = GetClockSec() - start_matrixmul32;
	printf("elapsed_matrixmul32=%.2fs\n", elapsed_matrixmul32);

	double start_int8 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		_MatrixMul(&inta, X_NOTRANS, &intb, X_NOTRANS, &intc);
	}
	double elapsed_int8 = GetClockSec() - start_int8;
	printf("elapsed_int8=%.2fs\n", elapsed_int8);


}

void TimeTest() {

	XMem * mem;
	int memSize = 1024;
	int devId = 0;
	int dim = 512;
	mem = new XMem(devId, FREE_ON_THE_FLY, (MTYPE)MILLION * 256, 1024, MILLION * 128);
	mem->SetDesiredSize(devId, 0, (MTYPE)memSize * MILLION);

	XTensor a;
	XTensor b;
	XTensor c;

	XTensor halfa;
	XTensor halfb;
	XTensor halfc;

	InitTensor2D(&a, dim, dim, X_FLOAT, devId, mem);
	InitTensor2D(&b, dim, dim, X_FLOAT, devId, mem);
	InitTensor2D(&c, dim, dim, X_FLOAT, devId, mem);

	InitTensor2D(&halfc, dim, dim, X_FLOAT16, devId, mem);

	a.SetDataRand(-1.0F, 1.0F);
	b.SetDataRand(-1.0F, 1.0F);

	halfa = ConvertDataType(a, X_FLOAT16);
	halfb = ConvertDataType(b, X_FLOAT16);

	int repeat = 100000;

	printf("=========================================\n");
	printf("test on sum\n");
	double start_sum32 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		c = Sum(&a, &b);
	}
	double elapsed_sum32 = GetClockSec() - start_sum32;
	printf("elapsed_sum32=%.2fs\n", elapsed_sum32);

	double start_sum16 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		halfc = Sum(&halfa, &halfb);
	}
	double elapsed_sum16 = GetClockSec() - start_sum16;
	printf("elapsed_sum16=%.2fs\n", elapsed_sum16);
	printf("=========================================\n");

	/*printf("test on sub\n");
	double start_sub32 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		c = Sub(&a, &b);
	}
	double elapsed_sub32 = GetClockSec() - start_sub32;
	printf("elapsed_sub32=%.2fs\n", elapsed_sub32);

	double start_sub16 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		halfc = Sub(&halfa, &halfb);
	}
	double elapsed_sub16 = GetClockSec() - start_sub16;
	printf("elapsed_sub16=%.2fs\n", elapsed_sub16);
	printf("=========================================\n");*/

	/*printf("test on div\n");
	double start_div32 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		c = Div(&a, &b);
	}
	double elapsed_div32 = GetClockSec() - start_div32;
	printf("elapsed_div32=%.2fs\n", elapsed_div32);

	double start_div16 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		halfc = Div(&halfa, &halfb);
	}
	double elapsed_div16 = GetClockSec() - start_div16;
	printf("elapsed_div16=%.2fs\n", elapsed_div16);
	printf("=========================================\n");*/

	/*printf("test on multiply\n");
	double start_multiply32 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		c = Multiply(&a, &b);
	}
	double elapsed_multiply32 = GetClockSec() - start_multiply32;
	printf("elapsed_multiply32=%.2fs\n", elapsed_multiply32);

	double start_multiply16 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		halfc = Multiply(&halfa, &halfb);
	}
	double elapsed_multiply16 = GetClockSec() - start_multiply16;
	printf("elapsed_multiply16=%.2fs\n", elapsed_multiply16);
	printf("=========================================\n");*/

	printf("test on scaleandshift\n");
	double start_scaleandshift32 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		c = ScaleAndShift(&a, 1, 0);
	}
	double elapsed_scaleandshift32 = GetClockSec() - start_scaleandshift32;
	printf("elapsed_scaleandshift32=%.2fs\n", elapsed_scaleandshift32);

	double start_scaleandshift16 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		halfc = ScaleAndShift(&halfa, 1, 0);
	}
	double elapsed_scaleandshift16 = GetClockSec() - start_scaleandshift16;
	printf("elapsed_scaleandshift16=%.2fs\n", elapsed_scaleandshift16);
	printf("=========================================\n");

	printf("test on reducesum\n");
	double start_reducesum32 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		c = ReduceSum(&a, 1);
	}
	double elapsed_reducesum32 = GetClockSec() - start_reducesum32;
	printf("elapsed_reducesum32=%.2fs\n", elapsed_reducesum32);

	double start_reducesum16 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		halfc = ReduceSum(&halfa, 1);
	}
	double elapsed_reducesum16 = GetClockSec() - start_reducesum16;
	printf("elapsed_reducesum16=%.2fs\n", elapsed_reducesum16);
	printf("=========================================\n");

	printf("test on reducemax\n");
	double start_reducemax32 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		c = ReduceMax(&a, 1);
	}
	double elapsed_reducemax32 = GetClockSec() - start_reducemax32;
	printf("elapsed_reducemax32=%.2fs\n", elapsed_reducemax32);

	double start_reducemax16 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		halfc = ReduceMax(&halfa, 1);
	}
	double elapsed_reducemax16 = GetClockSec() - start_reducemax16;
	printf("elapsed_reducemax16=%.2fs\n", elapsed_reducemax16);
	printf("=========================================\n");

	printf("test on logsoftmax\n");
	double start_logsoftmax32 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		c = LogSoftmax(&a, 1);
	}
	double elapsed_logsoftmax32 = GetClockSec() - start_logsoftmax32;
	printf("elapsed_logsoftmax32=%.2fs\n", elapsed_logsoftmax32);

	double start_logsoftmax16 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		halfc = LogSoftmax(&halfa, 1);
	}
	double elapsed_logsoftmax16 = GetClockSec() - start_logsoftmax16;
	printf("elapsed_logsoftmax16=%.2fs\n", elapsed_logsoftmax16);
	printf("=========================================\n");

	printf("test on matrixmul\n");
	double start_matrixmul32 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		c = MatrixMul(&a, &b);
	}
	double elapsed_matrixmul32 = GetClockSec() - start_matrixmul32;
	printf("elapsed_matrixmul32=%.2fs\n", elapsed_matrixmul32);

	double start_matrixmul16 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		halfc = MatrixMul(&halfa, &halfb);
	}
	double elapsed_matrixmul16 = GetClockSec() - start_matrixmul16;
	printf("elapsed_matrixmul16=%.2fs\n", elapsed_matrixmul16);
	printf("=========================================\n");
	printf("test on convert\n");
	double start_convert32to16 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		halfa = ConvertDataType(a, X_FLOAT16);
	}
	double elapsed_convert32to16 = GetClockSec() - start_convert32to16;
	printf("elapsed_convert32to16=%.2fs\n", elapsed_convert32to16);

	double start_convert16to32 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		a = ConvertDataType(halfa, X_FLOAT);
	}
	double elapsed_convert16to32 = GetClockSec() - start_convert16to32;
	printf("elapsed_convert16to32=%.2fs\n", elapsed_convert16to32);
	printf("=========================================\n");
	delete mem;
}

void MultiplyDimTime() {
	int memSize = 1024;
	int devId = 0;
	int dim1 = 1024;
	int dim2 = 32;
	XMem * mem;
	mem = new XMem(devId, FREE_ON_THE_FLY, (MTYPE)MILLION * 256, 1024, MILLION * 128);
	mem->SetDesiredSize(devId, 0, (MTYPE)memSize * MILLION);

	XTensor a;
	XTensor b;
	XTensor c;

	//XMem *mem = new XMem(devId, FREE_ON_THE_FLY, 128 * MILLION, 1024, 128 * MILLION);
	//mem->SetDesiredSize(0,0,memSize*MILLION);

	InitTensor2D(&a, dim1, dim1, X_FLOAT, devId);
	InitTensor2D(&b, dim2, dim2, X_FLOAT, devId);
	InitTensor2D(&c, dim1, dim1, X_FLOAT, devId);

	a.SetDataRand(-1.0F, 1.0F);
	b.SetDataRandn(-1.0F, 1.0F);

	int repeat = 2000;
	printf("test on MultiplyDim\n");

	double start = GetClockSec();
	for (int j = 0; j <= repeat; j++) {
		c = MultiplyDim(&a, &b, 0);
	}
	double elapsed = GetClockSec() - start;
	printf("elapsed_MultiplyDim32=%.4fs \n", elapsed);

	XTensor halfa;
	XTensor halfb;
	XTensor halfc;
	InitTensor2D(&halfc, dim1, dim1, X_FLOAT16, devId, mem);
	halfa = ConvertDataType(a, X_FLOAT16);
	halfb = ConvertDataType(b, X_FLOAT16);
	double starthalf = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		halfc = MultiplyDim(&halfa, &halfb, 0);
	}
	double elapsedhalf = GetClockSec() - starthalf;
	printf("elapsed_MultiplyDim16=%.4fs\n", elapsedhalf);
}

void TimeTestGemm() {
	XMem * mem;
	int memSize = 1024;
	delete mem;
	mem = new XMem(0, FREE_ON_THE_FLY, (MTYPE)MILLION * 256, 1024, MILLION * 128);
	mem->SetDesiredSize(0, 0, (MTYPE)memSize * MILLION);

	XTensor a;
	XTensor b;
	XTensor c;
	XTensor halfa;
	XTensor halfb;
	XTensor halfc;
	int dim1 = 512;
	int dim2 = 1024;
	//InitTensor3D(&a, 86, 48, 256, X_FLOAT, 0, mem);
	//InitTensor2D(&b, 256, 256, X_FLOAT, 0, mem);

	//InitTensor4D(&a, 8, 86, 48, 48, X_FLOAT, 0, mem);
	//InitTensor4D(&b, 8, 86, 48, 32, X_FLOAT, 0, mem);
	InitTensor2D(&a, dim1, dim2, X_FLOAT, 0, mem);
	InitTensor2D(&b, dim1, dim2, X_FLOAT, 0, mem);
	//InitTensor4D(&a, 8, 86, 48, 32, X_FLOAT, 0);
	//InitTensor4D(&b, 8, 86, 48, 32, X_FLOAT, 0);

	a.SetDataRand(-1.0F, 1.0F);
	b.SetDataRand(-1.0F, 1.0F);
	halfa = ConvertDataType(a, X_FLOAT16);
	halfb = ConvertDataType(b, X_FLOAT16);

	//a.Dump(&a, stderr, "a:", 10);
	//b.Dump(&b, stderr, "b:", 10);

	//halfa.Dump(&a, stderr, "halfa:", 10);
	//halfb.Dump(&b, stderr, "halfb:", 10);

	int repeat = 10000;
	printf("=========================================\n");
	double start_matrixmul16 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		halfc = BMMul(halfa, X_NOTRANS, halfb, X_TRANS);
	}
	double elapsed_matrixmul16 = GetClockSec() - start_matrixmul16;
	printf("elapsed_matrixmul16=%.4fs\n", elapsed_matrixmul16);
	printf("------------------------------------------\n");
	double start_matrixmul32 = GetClockSec();
	for (int i = 0; i < repeat; i++) {
		c = BMMul(a, X_NOTRANS, b, X_TRANS);
	}
	double elapsed_matrixmul32 = GetClockSec() - start_matrixmul32;
	printf("elapsed_matrixmul32=%.4fs\n", elapsed_matrixmul32);
	printf("=========================================\n");
	c.Dump(&c, stderr, "c:", 10);
	halfc.Dump(&halfc, stderr, "halfc:", 10);
}

void InitTensorFP16Test() {
    XTensor a;
    InitTensor2D(&a, 1, 10, X_FLOAT, 0);
    a.SetDataRand(-10.0F, 10.0F);
    XTensor halfA;
    halfA = ConvertDataType(a, X_FLOAT16);
    halfA.Dump(&halfA, stderr, "halfA:");

    XTensor b;
    InitTensor2D(&b, 1, 10, X_FLOAT16, 0);
    _SetDataRand(&b, -10.0F, 10.0F);
    b.Dump(&b, stderr, "b:");
}

void ScaleAndShiftFP16Test() {
    XTensor a;
    XTensor intA;
    XTensor b;
    XTensor intB;

    InitTensor2D(&a, 1, 10, X_FLOAT, 0);
    a.SetDataRand(-10.0F, 10.0F);

    a.Dump(stderr, "a:");

    intA = ConvertDataType(a, X_INT);

    intB = ScaleAndShift(intA, 2, 0);

    b = ConvertDataType(intB, X_FLOAT);

    b.Dump(stderr, "b:");
}

void ClipFP16Test() {
    XTensor a;
    XTensor intA;
    XTensor b;
    XTensor intB;

    InitTensor2D(&a, 1, 10, X_FLOAT, 0);
    a.SetDataRand(-10.0F, 10.0F);

    a.Dump(stderr, "a:");

    intA = ConvertDataType(a, X_INT);

    intB = Clip(intA, -1, 1);

    b = ConvertDataType(intB, X_FLOAT);

    b.Dump(stderr, "b:");
}

void LogSoftmaxFP16Test() {

    XTensor a;
    XTensor halfA;
    XTensor b;
    XTensor halfB;

    InitTensor3D(&a, 2, 2, 2, X_FLOAT, 0);
    a.SetDataRand(-1.0F, 1.0F);

    halfA = ConvertDataType(a, X_FLOAT16);

    b = LogSoftmax(a, 1);
    halfB = LogSoftmax(halfA, 1);

    b.Dump(stderr, "sum:");
    halfB.Dump(&halfB, stderr, "halfSum:");

}

void ReduceSumFP16Test()
{
    XTensor a;
    XTensor sum;
    XTensor halfA;
    XTensor halfSum;
    InitTensor2D(&a, 10, 10, X_FLOAT, 0);

    a.SetDataRand(-5.0F, 5.0F);

    halfA = ConvertDataType(a, X_FLOAT16);

    sum = ReduceSum(a, 1);
    halfSum = ReduceSum(halfA, 1);

    sum.Dump(stderr, "sum:");
    halfSum.Dump(&halfSum, stderr, "halfSum:");
}

void ReduceMaxFP16Test()
{
    XTensor a;
    XTensor max;
    XTensor halfA;
    XTensor halfMax;
    InitTensor2D(&a, 10, 10, X_FLOAT, 0);

    a.SetDataRand(-5.0F, 5.0F);

    halfA = ConvertDataType(a, X_FLOAT16);

    max = ReduceMax(a, 1);
    halfMax = ReduceMax(halfA, 1);

    max.Dump(stderr, "max:");
    halfMax.Dump(&halfMax, stderr, "halfMax:");
}


void HardTanHFP16Test()
{
    XTensor a;
    XTensor b;
    XTensor halfA;
    XTensor halfB;
    InitTensor2D(&a, 5, 5, X_FLOAT, 0);
    InitTensor2D(&b, 5, 5, X_FLOAT, 0);

    a.SetDataRand(-1.0F, 4.0F);
    b.SetDataRand(-1.0F, 4.0F);

    halfA = ConvertDataType(a, X_FLOAT16);
    halfB = ConvertDataType(b, X_FLOAT16);

    a.Dump(stderr, "a:");
    b.Dump(stderr, "b:");

    b = HardTanH(a);
    halfB = HardTanH(halfA);

    b.Dump(stderr, "b:");
    halfB.Dump(&halfB, stderr, "halfB:");

}

void GatherFP16Test() {
    XTensor a;
    XTensor b;
    XTensor srcIndex;
    XTensor halfA;
    XTensor halfB;
    XTensor c;

    InitTensor1D(&srcIndex, 2, X_INT, 0);
    int m = 0;
    int n = 1;
    srcIndex.Set1DInt(m, 0);
    srcIndex.Set1DInt(n, 1);

    InitTensor2D(&a, 3, 2, X_FLOAT, 0);
    InitTensor2D(&b, 2, 2, X_FLOAT, 0);
    InitTensor2D(&halfB, 2, 2, X_FLOAT16, 0);

    a.SetDataRand(-5.0F, 5.0F);
    halfA = ConvertDataType(a, X_FLOAT16);

    a.Dump(stderr, "a:");
    _Gather(&a, &b, &srcIndex);

    b.Dump(stderr, "b:");

    _Gather(&halfA, &halfB, &srcIndex);
    c = ConvertDataType(halfB, X_FLOAT);
    c.Dump(stderr, "c:");
}

void SumFP16Test()
{
    XTensor a;
    XTensor b;
    XTensor halfA;
    XTensor halfB;
    InitTensor2D(&a, 5, 5, X_FLOAT, 0);
    InitTensor2D(&b, 5, 5, X_FLOAT, 0);

    a.SetDataRand(-1.0F, 4.0F);
    b.SetDataRand(-1.0F, 4.0F);

    halfA = ConvertDataType(a, X_FLOAT16);
    halfB = ConvertDataType(b, X_FLOAT16);

    a.Dump(stderr, "a:");
    b.Dump(stderr, "b:");

    b = Sum(a, b, -0.4F);
    halfB = Sum(halfA, halfB, -0.4F);

    b.Dump(stderr, "b:");
    halfB.Dump(&halfB, stderr, "halfB:");

}

void ConvertDataTypeTest()
{
    int rnum = 0;
    for (int i = 0; i <= rnum; i++)
    {
        XTensor a;
        InitTensor2D(&a, 2, 2, X_FLOAT, 0);

        XTensor halfa;
        InitTensor2D(&halfa, 2, 2, X_FLOAT16, 0);

        XTensor a1;
        InitTensor2D(&a1, 2, 2, X_FLOAT, 0);

        a.SetDataRand(-10.0F, 10.0F);

        a.Dump(stderr, "a:");

        halfa = ConvertDataType(a, X_FLOAT16);

        a1 = ConvertDataType(halfa, X_FLOAT);

        a1.Dump(stderr, "halfa:");
    }
}

void ConvertDataTypeBackwardTest()
{
    int rnum = 0;
    for (int i = 0; i <= rnum; i++)
    {
        XTensor a;
        InitTensor2D(&a, 2, 2, X_FLOAT, 0);
        a.SetDataRand(2.0F, 2.0F);
        a.Dump(stderr, "a:");

        XTensor halfA;

        XTensor a1;

        halfA = ConvertDataType(a, X_FLOAT16);

        a1 = ConvertDataType(halfA, X_FLOAT);

        a1.grad = NewTensor(&a1);
        a1.grad->SetDataRand(3.0F, 3.0F);

        a1.grad->Dump(stderr, "a1.grad:");

        XNet testBackward;

        printf("1");

        testBackward.Backward(a1);

        printf("2");

        halfA.grad->Dump(stderr, "halfA.grad:");
        a.grad->Dump(stderr, "a.grad:");
        
    }
}

XTensor * stack(XList& list, int leadingDim)
{
    size_t size = list.count;
    if (list.count == 0)
        return NULL;
    XTensor * sample = (XTensor*)list.Get(0);

    XTensor merge_tensor;
    int order = sample->order;
    int * dim = new int[order];
    for (int i = 0; i < order; i++)
        dim[i] = sample->GetDim(i);
    dim[leadingDim] *= size;

    InitTensor(&merge_tensor, order, dim, DEFAULT_DTYPE, sample->denseRatio, sample->devID, sample->mem);

    _Merge(&list, &merge_tensor, leadingDim);
    delete[] dim;

    order += 1;
    dim = new int[order];
    dim[0] = size;
    for (size_t i = 1; i < order; i++) {
        if (i != leadingDim)
            dim[i] = sample->GetDim(i-1);
        else
            dim[i] = sample->GetDim(i-1)/size;
    }

    XTensor * split_tensor = new XTensor(order, dim, DEFAULT_DTYPE, sample->denseRatio, sample->devID, sample->mem);
    _Split(&merge_tensor, split_tensor, leadingDim, size);
    delete[] dim;

    return split_tensor;
}

void xcTest()
{
    int * dimSize = new int[2]; 
    dimSize[0] = 2;
    dimSize[1] = 4;

    XTensor t1;
    InitTensor2D(&t1, 2, 4, X_FLOAT, 0, NULL);
    XTensor t2;
    InitTensor2D(&t2, 2, 4, X_FLOAT, 0, NULL);
    XTensor tensor;
    
    _SetDataFixed(&t1, 1.0F);
    _SetDataFixed(&t2, 2.0F);

    tensor = t1 + t2;

    XList smalls;

    XTensor first;
    XTensor second;
    InitTensor2D(&first, 2, 2, X_FLOAT, 0, NULL);
    InitTensor2D(&second, 2, 2, X_FLOAT, 0, NULL);
    smalls.Add(&t1);
    smalls.Add(&t2);

    XTensor* result = stack(smalls, 0);
    result->Dump(stderr, "", 100);
}

void BackwardTest()
{
    XNet net;

    XTensor a;
    XTensor b;
    XTensor c;
    XTensor mean;
    XTensor origin;
    InitTensor2D(&a, 2, 3);
    InitTensor1D(&b, 2);

    a.SetZeroAll();
    b.SetZeroAll();
    a.Set2D(1.0F, 0, 0);
    a.Set2D(2.0F, 0, 1);
    a.Set2D(3.0F, 0, 2);
    a.Set2D(4.0F, 1, 0);
    a.Set2D(5.0F, 1, 1);
    a.Set2D(6.0F, 1, 2);

    b.Set1D(2.0F, 0);
    b.Set1D(1.0F, 1);

    c = DivDim(a, b, 0);
    c.Dump(stderr, "c:");

    //XLink::ShowNetwork(stderr, &c);

    net.Backward(c);

    net.Dump(stderr);

}

void TransposeTest()
{
#ifdef USE_CUDA
    XMem mem0(0, UNI_FREE, MILLION * 64, 1024, MILLION * 64);
    //XMem mem1(1, UNI_FREE, MILLION * 64, 1024, MILLION * 64);
    XTensor x;
    XTensor y;
    XTensor z;

    int loops = 2000;

    int B = 3 * 2 * 4;
    int K = 8 * 1;
    int N = 50;
    int H = 512 * 4;

    int nnn = GDevs.nGPU;

    InitTensor3D(&x, B, N, H, X_FLOAT, 0);
    InitTensor4D(&y, K, B, N, H/K, X_FLOAT, 0);
    InitTensor3D(&z, B, N, H, X_FLOAT, 0);

    cudaEvent_t ctime0;
    cudaEvent_t ctime1;
    cudaEvent_t ctime2;
    cudaEvent_t ctime3;
    cudaEvent_t ctime4;
    cudaEvent_t ctime5;

    float elapsedSplit = 0.0;
    float elapsedMerge = 0.0;
    float elapsedSum = 0.0;

    cudaEventCreate(&ctime0);
    cudaEventCreate(&ctime1);
    cudaEventCreate(&ctime2);
    cudaEventCreate(&ctime3);
    cudaEventCreate(&ctime4);
    cudaEventCreate(&ctime5);

    cudaEventRecord(ctime0, 0);

    double time0 = GetClock();
    for(int i = 0; i < loops; i++)
        _Split(&x, &y, 2, K);
    double time1 = GetClock();
    
    cudaEventRecord(ctime1, 0);
    cudaEventSynchronize(ctime1);
    cudaEventElapsedTime(&elapsedSplit, ctime0, ctime1);

    cudaEventRecord(ctime2, 0);

    double time2 = GetClock();
    for(int i = 0; i < loops; i++)
        _Merge(&y, &x, 3);
    double time3 = GetClock();

    cudaEventRecord(ctime3, 0);
    cudaEventSynchronize(ctime3);
    cudaEventElapsedTime(&elapsedMerge, ctime2, ctime3);

    cudaEventRecord(ctime4, 0);

    double time4 = GetClock();
    for(int i = 0; i < loops; i++)
        _Sum(&x, &z, &x);
    double time5 = GetClock();

    cudaEventRecord(ctime5, 0);
    cudaEventSynchronize(ctime5);
    cudaEventElapsedTime(&elapsedSum, ctime4, ctime5);

    fprintf(stderr, "split:%f merge:%f sum:%f\n", time1 - time0, time3 - time2, time5 - time4);
    fprintf(stderr, "split:%f merge:%f sum:%f\n", elapsedSplit, elapsedMerge, elapsedSum);
#endif
}

void SumDimTest()
{
    XTensor x;
    XTensor y;
    XTensor z;

    int a = 5;
    int b = 7;
    int c = 3;

    InitTensor3D(&x, a, b, c, X_FLOAT, -1);
    InitTensor1D(&y, c, X_FLOAT, -1);
    InitTensor3D(&z, a, b, c, X_FLOAT, -1);

    x.SetZeroAll();
    y.SetZeroAll();
    z.SetZeroAll();

    DTYPE * data = new DTYPE[x.unitNum];

    for(int i = 0; i < x.unitNum; i++)
        data[i] = (DTYPE)i;
    x.SetData(data, x.unitNum);

    for(int i = 0; i < y.unitNum; i++)
        data[i] = -(DTYPE)i;
    y.SetData(data, y.unitNum);

    _SumDim(&x, &y, &z, 2);

    z.Dump(stderr, "z:");

    delete[] data;
}

void SplitBackwardTest()
{
    int * dimSize = new int[2];
    dimSize[0] = 2;
    dimSize[1] = 4;

    XTensor t1;
    InitTensor2D(&t1, 2, 4, X_FLOAT, 0, NULL);
    XTensor t2;
    InitTensor2D(&t2, 2, 4, X_FLOAT, 0, NULL);
    XTensor tensor;
    
    //_SetDataFixedFloat(&t1, 1.0F);
    //_SetDataFixedFloat(&t2, 2.0F);
    t1.SetDataRand();
    t2.SetDataRand();

    tensor = t1 + t2;

    XList smalls;

    XTensor first;
    XTensor second;
    InitTensor2D(&first, 2, 2, X_FLOAT, 0, NULL);
    InitTensor2D(&second, 2, 2, X_FLOAT, 0, NULL);
    smalls.Add(&first);
    smalls.Add(&second);

    Split(tensor, smalls, 1, 2);

    XTensor mul;
    mul = Sum(first, second);

    XNet net;
    net.Backward(mul);
    net.Dump(stderr);

    printf("Done!");
}

void MemTest()
{
    XMem * mem;
    mem = new XMem(0, FREE_ON_THE_FLY, (MTYPE)MILLION, 1024, MILLION);
    
    XTensor tensor;
    InitTensor2D(&tensor, 2, 4, X_FLOAT, 0, mem);
    
    tensor.SetZeroAll();

    tensor.Dump(stderr);

    delete mem;

    if (tensor.mem != NULL) {
        printf("It isn't null!\n");
        printf("%d\n", (int)tensor.mem->signature);
    }
    else {
        printf("It's null\n");
    }
    tensor.Dump(stderr);
}